require 'childprocess'

module Graal::Backend::Exe

   class Git

      COMMIT_SHOW_FORMAT = '--pretty=format:' + [
         '%P',
         '%an', '%ae', '%at',
         '%cn', '%ce', '%ct',
         '%s', '%-b'
      ] * '%n'

      class Result

         attr_reader :exit_code, :output, :error

         def initialize(exit_code, output_io, error_io)
            @exit_code = exit_code
            @output = output_io.read
            output_io.close
            @error = error_io.read
            error_io.read
         end

         def error?
            @exit_code != 0
         end
      end

      def initialize(repository)
         @repository = repository
      end

      def run(cmd, *args)
         do_run(['git', cmd, *args])
      end

      def run_with_pathspec(cmd, *args, pathspec)
         command_line = ['git', cmd, *args]
         unless pathspec.nil? or pathspec.empty?
            command_line << '--'
            command_line << pathspec
         end
         do_run(command_line)
      end

      def heads
         result = run('branch', '--no-color', '--no-column')
         result.output.gsub(/^[ *]*/, '').split("\n")
      end

      def tags
         result = run('tag', '--no-column')
         result.output.split("\n")
      end

      def gitid_for(revision, path)
         if path.empty?
            result = run('rev-parse', revision).output.chomp
         else
            result = run('ls-tree', revision, path)
            result.output.split(' ')[2]
         end
      end

      def commit_info(revision)
         result = run('show', '--quiet', COMMIT_SHOW_FORMAT, revision)
         ary = result.output.split("\n")
         ary[8] = ary[8..-1] * "\n" if ary.size > 9
         ary
      end

      def log(revision, path, raw_opts = {})
         opts = parse_opts(raw_opts)
         result = run_with_pathspec('rev-list', *opts, revision, path)
         result.output.split("\n")
      end

      def child(gitid, fullpath)
         result = run('ls-tree', gitid, fullpath)
         return result.output.split(" ")[1..2]
      end

      def children(gitid, path, filter = nil)
         result = run('ls-tree', gitid, path)
         children = []
         result.output.split("\n").each do |line|
            splitted = line.split(" ")
            next unless filter.nil? or splitted[1] == filter
            children << [splitted[1], splitted[3], splitted[2]]
         end
         children
      end

      def blob_contents(gitid)
         run('cat-file', '-p', gitid).output
      end

   private

      def parse_opts(opts)
         opts.each_with_object([]) do |opt, array|
            array << '--' + (opt[1] == true ? opt[0] : opt * '=')
         end
      end

      def do_run(command_line)
         process = ChildProcess.build(*command_line)
         process.cwd = @repository
         out, process.io.stdout, err, process.io.stderr = *IO.pipe, *IO.pipe

         process.start
         process.wait
         process.io.stdout.close
         process.io.stderr.close

         Result.new(process.exit_code, out, err)
      end
   end
end
